'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var common = require('@ts-morph/common');

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __decorate(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
}

class SourceFileCache {
    constructor(fileSystemWrapper, compilerOptions) {
        this.fileSystemWrapper = fileSystemWrapper;
        this.compilerOptions = compilerOptions;
        this.sourceFilesByFilePath = new Map();
        this.documentRegistry = new common.DocumentRegistry(fileSystemWrapper);
    }
    containsSourceFileAtPath(filePath) {
        return this.sourceFilesByFilePath.has(filePath);
    }
    getSourceFilePaths() {
        return this.sourceFilesByFilePath.keys();
    }
    getSourceFiles() {
        return this.sourceFilesByFilePath.values();
    }
    getSourceFileVersion(sourceFile) {
        return this.documentRegistry.getSourceFileVersion(sourceFile);
    }
    getSourceFileFromCacheFromFilePath(filePath) {
        return this.sourceFilesByFilePath.get(filePath);
    }
    addOrGetSourceFileFromFilePath(filePath, options) {
        let sourceFile = this.sourceFilesByFilePath.get(filePath);
        if (sourceFile == null && this.fileSystemWrapper.fileExistsSync(filePath)) {
            sourceFile = this.createSourceFileFromText(filePath, this.fileSystemWrapper.readFileSync(filePath, this.compilerOptions.getEncoding()), options);
        }
        return sourceFile;
    }
    createSourceFileFromText(filePath, text, options) {
        filePath = this.fileSystemWrapper.getStandardizedAbsolutePath(filePath);
        const hasBom = common.StringUtils.hasBom(text);
        if (hasBom)
            text = common.StringUtils.stripBom(text);
        const sourceFile = this.documentRegistry.createOrUpdateSourceFile(filePath, this.compilerOptions.get(), common.ts.ScriptSnapshot.fromString(text), options.scriptKind);
        this.setSourceFile(sourceFile);
        return sourceFile;
    }
    setSourceFile(sourceFile) {
        const standardizedFilePath = this.fileSystemWrapper.getStandardizedAbsolutePath(sourceFile.fileName);
        sourceFile.fileName = standardizedFilePath;
        this.documentRegistry.updateDocument(standardizedFilePath, this.compilerOptions.get(), common.ts.ScriptSnapshot.fromString(sourceFile.text), this.getSourceFileVersion(sourceFile), sourceFile["scriptKind"]);
        this.fileSystemWrapper.queueMkdir(common.FileUtils.getDirPath(standardizedFilePath));
        this.sourceFilesByFilePath.set(standardizedFilePath, sourceFile);
    }
    removeSourceFile(filePath) {
        this.sourceFilesByFilePath.delete(filePath);
    }
    containsDirectoryAtPath(dirPath) {
        return this.fileSystemWrapper.directoryExistsSync(dirPath);
    }
    getChildDirectoriesOfDirectory(dirPath) {
        return this.fileSystemWrapper.getDirectories(dirPath);
    }
}

class Project {
    constructor(options = {}) {
        verifyOptions();
        this.fileSystem = getFileSystem();
        this._fileSystemWrapper = new common.TransactionalFileSystem(this.fileSystem);
        const tsConfigResolver = options.tsConfigFilePath == null
            ? undefined
            : new common.TsConfigResolver(this._fileSystemWrapper, this._fileSystemWrapper.getStandardizedAbsolutePath(options.tsConfigFilePath), getEncodingFromProvidedOptions());
        const tsCompilerOptions = getCompilerOptions();
        this.compilerOptions = new common.CompilerOptionsContainer();
        this.compilerOptions.set(tsCompilerOptions);
        this._sourceFileCache = new SourceFileCache(this._fileSystemWrapper, this.compilerOptions);
        const resolutionHost = !options.resolutionHost
            ? undefined
            : options.resolutionHost(this.getModuleResolutionHost(), () => this.compilerOptions.get());
        const newLineKind = "\n";
        const { languageServiceHost, compilerHost } = common.createHosts({
            transactionalFileSystem: this._fileSystemWrapper,
            sourceFileContainer: this._sourceFileCache,
            compilerOptions: this.compilerOptions,
            getNewLine: () => newLineKind,
            resolutionHost: resolutionHost || {}
        });
        this.languageServiceHost = languageServiceHost;
        this.compilerHost = compilerHost;
        if (tsConfigResolver != null && options.addFilesFromTsConfig !== false) {
            this._addSourceFilesForTsConfigResolver(tsConfigResolver, tsCompilerOptions);
            if (!options.skipFileDependencyResolution)
                this.resolveSourceFileDependencies();
        }
        function verifyOptions() {
            if (options.fileSystem != null && options.useInMemoryFileSystem)
                throw new common.errors.InvalidOperationError("Cannot provide a file system when specifying to use an in-memory file system.");
            if (options.skipLoadingLibFiles && !options.useInMemoryFileSystem) {
                throw new common.errors.InvalidOperationError(`The ${"skipLoadingLibFiles"} option can only be true when ${"useInMemoryFileSystem"} is true.`);
            }
        }
        function getFileSystem() {
            var _a;
            if (options.useInMemoryFileSystem)
                return new common.InMemoryFileSystemHost({ skipLoadingLibFiles: options.skipLoadingLibFiles });
            return _a = options.fileSystem, (_a !== null && _a !== void 0 ? _a : new common.RealFileSystemHost());
        }
        function getCompilerOptions() {
            return Object.assign(Object.assign({}, getTsConfigCompilerOptions()), (options.compilerOptions || {}));
        }
        function getTsConfigCompilerOptions() {
            if (tsConfigResolver == null)
                return {};
            return tsConfigResolver.getCompilerOptions();
        }
        function getEncodingFromProvidedOptions() {
            const defaultEncoding = "utf-8";
            if (options.compilerOptions != null)
                return options.compilerOptions.charset || defaultEncoding;
            return defaultEncoding;
        }
    }
    addSourceFileAtPathIfExists(filePath, options) {
        return this._sourceFileCache.addOrGetSourceFileFromFilePath(this._fileSystemWrapper.getStandardizedAbsolutePath(filePath), {
            scriptKind: options && options.scriptKind
        });
    }
    addSourceFileAtPath(filePath, options) {
        const sourceFile = this.addSourceFileAtPathIfExists(filePath, options);
        if (sourceFile == null)
            throw new common.errors.FileNotFoundError(this._fileSystemWrapper.getStandardizedAbsolutePath(filePath));
        return sourceFile;
    }
    addSourceFilesByPaths(fileGlobs) {
        if (typeof fileGlobs === "string")
            fileGlobs = [fileGlobs];
        const sourceFiles = [];
        for (const filePath of this._fileSystemWrapper.globSync(fileGlobs)) {
            const sourceFile = this.addSourceFileAtPathIfExists(filePath);
            if (sourceFile != null)
                sourceFiles.push(sourceFile);
        }
        return sourceFiles;
    }
    addSourceFilesFromTsConfig(tsConfigFilePath) {
        const standardizedFilePath = this._fileSystemWrapper.getStandardizedAbsolutePath(tsConfigFilePath);
        const resolver = new common.TsConfigResolver(this._fileSystemWrapper, standardizedFilePath, this.compilerOptions.getEncoding());
        return this._addSourceFilesForTsConfigResolver(resolver, resolver.getCompilerOptions());
    }
    createSourceFile(filePath, sourceFileText, options) {
        return this._sourceFileCache.createSourceFileFromText(this._fileSystemWrapper.getStandardizedAbsolutePath(filePath), sourceFileText || "", { scriptKind: options && options.scriptKind });
    }
    updateSourceFile(filePathOrSourceFile, sourceFileText, options) {
        if (typeof filePathOrSourceFile === "string")
            return this.createSourceFile(filePathOrSourceFile, sourceFileText, options);
        incrementVersion(filePathOrSourceFile);
        ensureScriptSnapshot(filePathOrSourceFile);
        return this._sourceFileCache.setSourceFile(filePathOrSourceFile);
        function incrementVersion(sourceFile) {
            let version = sourceFile.version || "-1";
            const parsedVersion = parseInt(version, 10);
            if (isNaN(parsedVersion))
                version = "0";
            else
                version = (parsedVersion + 1).toString();
            sourceFile.version = version;
        }
        function ensureScriptSnapshot(sourceFile) {
            if (sourceFile.scriptSnapshot == null)
                sourceFile.scriptSnapshot = common.ts.ScriptSnapshot.fromString(sourceFile.text);
        }
    }
    removeSourceFile(filePathOrSourceFile) {
        this._sourceFileCache.removeSourceFile(this._fileSystemWrapper.getStandardizedAbsolutePath(typeof filePathOrSourceFile === "string" ? filePathOrSourceFile : filePathOrSourceFile.fileName));
    }
    resolveSourceFileDependencies() {
        this.createProgram();
    }
    _addSourceFilesForTsConfigResolver(tsConfigResolver, compilerOptions) {
        const paths = tsConfigResolver.getPaths(compilerOptions);
        const addedSourceFiles = paths.filePaths.map(p => this.addSourceFileAtPath(p));
        return addedSourceFiles;
    }
    createProgram(options) {
        const oldProgram = this._oldProgram;
        const program = common.ts.createProgram(Object.assign({ rootNames: Array.from(this._sourceFileCache.getSourceFilePaths()), options: this.compilerOptions.get(), host: this.compilerHost, oldProgram }, options));
        this._oldProgram = program;
        return program;
    }
    getLanguageService() {
        return common.ts.createLanguageService(this.languageServiceHost, this._sourceFileCache.documentRegistry);
    }
    getSourceFileOrThrow(fileNameOrSearchFunction) {
        const sourceFile = this.getSourceFile(fileNameOrSearchFunction);
        if (sourceFile != null)
            return sourceFile;
        if (typeof fileNameOrSearchFunction === "string") {
            const fileNameOrPath = common.FileUtils.standardizeSlashes(fileNameOrSearchFunction);
            if (common.FileUtils.pathIsAbsolute(fileNameOrPath) || fileNameOrPath.indexOf("/") >= 0) {
                const errorFileNameOrPath = this._fileSystemWrapper.getStandardizedAbsolutePath(fileNameOrPath);
                throw new common.errors.InvalidOperationError(`Could not find source file in project at the provided path: ${errorFileNameOrPath}`);
            }
            else {
                throw new common.errors.InvalidOperationError(`Could not find source file in project with the provided file name: ${fileNameOrSearchFunction}`);
            }
        }
        else {
            throw new common.errors.InvalidOperationError(`Could not find source file in project based on the provided condition.`);
        }
    }
    getSourceFile(fileNameOrSearchFunction) {
        const filePathOrSearchFunction = getFilePathOrSearchFunction(this._fileSystemWrapper);
        if (isStandardizedFilePath(filePathOrSearchFunction)) {
            return this._sourceFileCache.getSourceFileFromCacheFromFilePath(filePathOrSearchFunction);
        }
        const allSoureFilesIterable = this.getSourceFiles();
        return selectSmallestDirPathResult(function* () {
            for (const sourceFile of allSoureFilesIterable) {
                if (filePathOrSearchFunction(sourceFile))
                    yield sourceFile;
            }
        }());
        function getFilePathOrSearchFunction(fileSystemWrapper) {
            if (fileNameOrSearchFunction instanceof Function)
                return fileNameOrSearchFunction;
            const fileNameOrPath = common.FileUtils.standardizeSlashes(fileNameOrSearchFunction);
            if (common.FileUtils.pathIsAbsolute(fileNameOrPath) || fileNameOrPath.indexOf("/") >= 0)
                return fileSystemWrapper.getStandardizedAbsolutePath(fileNameOrPath);
            else
                return def => common.FileUtils.pathEndsWith(def.fileName, fileNameOrPath);
        }
        function selectSmallestDirPathResult(results) {
            let result;
            for (const sourceFile of results) {
                if (result == null || common.FileUtils.getDirPath(sourceFile.fileName).length < common.FileUtils.getDirPath(result.fileName).length)
                    result = sourceFile;
            }
            return result;
        }
        function isStandardizedFilePath(obj) {
            return typeof obj === "string";
        }
    }
    getSourceFiles() {
        return Array.from(this._sourceFileCache.getSourceFiles());
    }
    formatDiagnosticsWithColorAndContext(diagnostics, opts = {}) {
        return common.ts.formatDiagnosticsWithColorAndContext(diagnostics, {
            getCurrentDirectory: () => this._fileSystemWrapper.getCurrentDirectory(),
            getCanonicalFileName: fileName => fileName,
            getNewLine: () => opts.newLineChar || require("os").EOL
        });
    }
    getModuleResolutionHost() {
        return common.createModuleResolutionHost({
            transactionalFileSystem: this._fileSystemWrapper,
            getEncoding: () => this.compilerOptions.getEncoding(),
            sourceFileContainer: this._sourceFileCache
        });
    }
}
__decorate([
    common.Memoize
], Project.prototype, "getLanguageService", null);
__decorate([
    common.Memoize
], Project.prototype, "getModuleResolutionHost", null);

Object.defineProperty(exports, 'CompilerOptionsContainer', {
    enumerable: true,
    get: function () {
        return common.CompilerOptionsContainer;
    }
});
Object.defineProperty(exports, 'InMemoryFileSystemHost', {
    enumerable: true,
    get: function () {
        return common.InMemoryFileSystemHost;
    }
});
Object.defineProperty(exports, 'SettingsContainer', {
    enumerable: true,
    get: function () {
        return common.SettingsContainer;
    }
});
Object.defineProperty(exports, 'ts', {
    enumerable: true,
    get: function () {
        return common.ts;
    }
});
exports.Project = Project;
